/*
 * PROJECT: Indiana University Stiquito Colony Simulator and Control Software
 *
 * (c) Copyright 1996 Jason L. Almeter, Indiana University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by Indiana University '' in the documentation or other
 * materials provided with the distribution and in all advertising
 * materials mentioning features or use of this software. The
 * name of the University may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * MODULE:
 *	Robot Controllers
 *
 * FILE:
 *	wander.c
 *
 * ABSTRACT:
 *	avoids walls and robots
 *
 * $Revision: 1.1 $
 * $Date: 1996/01/18 17:11:15 $
 * $Author: jla $
 *
 * HISTORY:
 *
 * $Log: wander.c,v $
 * Revision 1.1  1996/01/18  17:11:15  jla
 * Initial revision
 *
 * Revision 1.2  1996/01/17  19:18:56  jla
 * Updated Cogyright verbage for release.
 *
 * Revision 1.1  1996/01/16  18:54:05  jla
 * Initial revision
 *
 * Revision 1.2  1996/01/15  14:49:26  jla
 * Enrolling in version control.
 *
 *
 */


#include <stdio.h>
#include <math.h>
#include "robot.h"

#define DEBUG	1

#define PI	3.14159265359

#define FORWARD	1.0/12.0	/* +-FORWARD*PI goes forward */
#define F_TURN	4.0/6.0		/* these forward turn */
                                /* rest reverse turn */

#define QUIVER		0.1

/*
 * define wall repulsion
 */
#define WALL_ZONE	(ROBOT_RADIUS+ROBOT_RADIUS)
#define WALL_FORCE	1.0

/*
 * define robot-robot repulsion
 */
#define ROBOT_ZONE	(3.0*ROBOT_RADIUS)
#define ROBOT_FORCE	2.0


/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: int vecToCommand(double x, double y)
 *
 * DESCRIPTION:
 *	translates a desired direction into a motor command
 *	assumes vector is relative to robot
 *
\*--------------------------------------------------------------------------*/


int vecToCommand(double x, double y)
{
   double theta;

   if(x == 0.0 && y == 0.0) {
      if(DEBUG) printf("%d stop\n",robotNumber);
      return CMD_STOP;
   }

   theta = atan2(y,x);

   if(fabs(theta) < PI*FORWARD) {
      if(DEBUG) printf("%d go forward\n",robotNumber);
      return CMD_F;
   }

   if(theta < 0.0 && theta > -F_TURN*PI) {
      if(DEBUG) printf("%d go forward right\n",robotNumber);
      return CMD_FR;
   }
   
   if(theta > 0.0 && theta < F_TURN*PI) {
      if(DEBUG) printf("%d go forward left\n",robotNumber);
      return CMD_FL;
   }
   
   if(theta > F_TURN*PI) {
      if(DEBUG) printf("%d go reverse right\n",robotNumber);
      return CMD_RR;
   }
   
   if(theta < -F_TURN*PI) {
      if(DEBUG) printf("%d go reverse left\n",robotNumber);
      return CMD_RL;
   }

   fprintf(stderr, "%d what the hell?\n",robotNumber);
/*   exit(-1);/**/
   return CMD_STOP;
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void robot_controller()
 *
 * DESCRIPTION:
 *	give a robot its directions
 *
\*--------------------------------------------------------------------------*/


void robot_controller(void)
{
   int i,forward;
   double dx,dy;
   double tx,ty;
   double norm;
   double scale;

   tx = 0.0;
   ty = 0.0;

   forward = 1;

   /* look for walls */
   if(fabs(NORTH_WALL - location[robotNumber].y) < WALL_ZONE) {
      scale = (WALL_ZONE - fabs(NORTH_WALL - location[robotNumber].y))/WALL_ZONE;
      ty -= WALL_FORCE*scale;
      forward = 0;
   }
   if(fabs(SOUTH_WALL - location[robotNumber].y) < WALL_ZONE) {
      scale = (WALL_ZONE - fabs(SOUTH_WALL - location[robotNumber].y))/WALL_ZONE;
      ty += WALL_FORCE*scale;
      forward = 0;
   }
   if(fabs(WEST_WALL - location[robotNumber].x) < WALL_ZONE) {
      scale = (WALL_ZONE - fabs(WEST_WALL - location[robotNumber].x))/WALL_ZONE;
      tx += WALL_FORCE*scale;
      forward = 0;
   }
   if(fabs(EAST_WALL - location[robotNumber].x) < WALL_ZONE) {
      scale = (WALL_ZONE - fabs(EAST_WALL - location[robotNumber].x))/WALL_ZONE;
      tx -= WALL_FORCE*scale;
      forward = 0;
   }
   /* convert to robot local coords */
   ToLocal(tx, ty, &dx, &dy);

   /* inter robot forces */
   for(i=0; i<MAX_ROBOTS; i++) {
      if(i != robotNumber) {
	 if(location[i].x < EAST_WALL &&
	    location[i].x > WEST_WALL &&
	    location[i].y < NORTH_WALL &&
	    location[i].y > SOUTH_WALL) {
	    tx = location[i].x - location[robotNumber].x;
	    ty = location[i].y - location[robotNumber].y;
	    norm = hypot(tx,ty);
	    if(norm < ROBOT_ZONE) {
	       double lx, ly;
	       scale = (ROBOT_ZONE - norm)/ROBOT_ZONE;
/*	       printf("too close %d %f\n",i,scale);/**/
	       tx /= norm;
	       ty /= norm;
	       ToLocal(tx, ty, &lx, &ly);
	       dx += -scale*ROBOT_FORCE*lx;
	       dy += -scale*ROBOT_FORCE*ly;
	       forward = 0;
	    }
	 }
      }
   }

   /* forward */
   if(forward) {
      if(DEBUG) printf("...ROAM\n");
      dx += 2.0*QUIVER;
   }

   SetCommand(vecToCommand(dx,dy));
}
