/*
 * PROJECT: Indiana University Stiquito Colony Simulator and Control Software
 *
 * (c) Copyright 1996 Jason L. Almeter, Indiana University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by Indiana University '' in the documentation or other
 * materials provided with the distribution and in all advertising
 * materials mentioning features or use of this software. The
 * name of the University may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * MODULE:
 *	Robot controller/simulator.
 *
 * FILE:
 *	motorsim.c
 *
 * ABSTRACT:
 *	Handle robot locomotor requests.
 *	Simulate robot motion.
 *	Send updated locations to vision simulator.
 *
 * $Revision: 1.4 $
 * $Date: 1996/01/17 19:02:14 $
 * $Author: jla $
 *
 * HISTORY:
 *
 * $Log: motorsim.c,v $
 * Revision 1.4  1996/01/17  19:02:14  jla
 * Updated Copyright verbage in preparation for release.
 *
 * Revision 1.3  1996/01/16  20:01:09  jla
 * Added minimal access control to robots.
 *
 * Revision 1.2  1996/01/16  17:11:06  jla
 * Changed robot 'command' into 'inform'.
 * This seems to have fixed the central explosion problem.  Central was
 * warning about issuing 'command' from 'broadcast'.  It meant business.
 * It is possible that the robots could swamp the motorsim module, but I
 * think it is under control.
 *
 * Revision 1.1  1996/01/15  14:22:35  jla
 * Initial revision
 *
 *
 */

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <math.h>
#include "tca.h"
#include "motor.h"
#include "motorsim.h"
#include "../Vision/visionsim.h"
#include "../Utils/random.h"
#include "../Clock/clock.h"

#define MODULE	"MotorSim Module"

typedef struct{		/* record current state of robot simulation	*/
   int cmd;
   double x;
   double y;
   double dX;
   double dY;
} StateT;

int cookie[MAX_ROBOTS];
int newCookie;

#define ToRad	(-3.1415926/180.0)

StateT state[MAX_ROBOTS];	/* record current state of colony sim	*/
CoordArray coord;		/* struct for sending update to vision	*/

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void incarnateQueryHnd(TCA_REF_PTR ref, IdentPtr id)
 *
 * DESCRIPTION:
 *	check the credentials of an incarnation request and handle it
 *	if OK
 *
\*--------------------------------------------------------------------------*/

void incarnateQueryHnd(TCA_REF_PTR ref, IdentPtr id)
{
	int num;
	CookieT reply;

	/* we like everyone */
	/* strcmp(id.ident, ...);
	 * strcmp(id.pass, ...);
	 */
	if(Success != tcaQuery(NumberQuery, NULL, &num)) {
		fprintf(stderr, "number query failed.\n");
		exit(-1);
	}
	cookie[num] = newCookie++;
	reply.number = num;
	reply.cookie = cookie[num];

	tcaReply(ref, &reply);
	tcaFreeData(tcaReferenceName(ref), id);
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void robotCmdHnd(TCA_REF_PTR ref, RobotCmdPtr cmd)
 *
 * DESCRIPTION:
 *	Enter a command for a robot into simulation state.
 *
\*--------------------------------------------------------------------------*/


void robotCmdHnd(TCA_REF_PTR ref, RobotCmdPtr cmd)
{
   int i;

   if(cmd->robot >= MAX_ROBOTS || cmd->robot < 0) {
      fprintf(stderr, "robot %d out of range (0-%d)\n", cmd->robot, MAX_ROBOTS);
      tcaFreeData(tcaReferenceName(ref), cmd);
      exit(-1);
      return;
   }
   if(cookie[cmd->robot] != cmd->cookie) {
	fprintf(stderr, "Bad cookie for robot %d.\n", cmd->robot);
	tcaFreeData(tcaReferenceName(ref), cmd);
	return;
   }
   state[cmd->robot].cmd = cmd->cmd;
   tcaFreeData(tcaReferenceName(ref), cmd);
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void simulate(int usec)
 *
 * DESCRIPTION:
 *	Perform the simulation.
 *	Disallow collision with walls or other robots.
 *	Adds a random factor to each movement.
 *
\*--------------------------------------------------------------------------*/


void simulate(int usec)
{
   int i,j,flag,turn;
   double norm,scale,tx,ty,dx,dy;
   XYType msg;

   scale = ((double) usec)/1.0e3;

   for(i=0; i<MAX_ROBOTS; i++) {
      flag = 0;
      turn = 0;
      /*
       * prepare robot position update
       */
      switch(state[i].cmd) {

       case CMD_F:
/*	 printf("%d F\n",i);/**/
	 dx = +state[i].dX*SPEED*scale;
	 dy = +state[i].dY*SPEED*scale;
	 flag = 1;
	 break;

       case CMD_FR:
/*	 printf("%d FR\n",i);/**/
	 turn = 1;
	 tx = state[i].dX*cos(ToRad*TURN_RADIUS*scale) -
	    state[i].dY*sin(ToRad*TURN_RADIUS*scale);
	 ty = state[i].dX*sin(ToRad*TURN_RADIUS*scale) +
	    state[i].dY*cos(ToRad*TURN_RADIUS*scale);
	 norm = sqrt(tx*tx+ty*ty);
	 tx /= norm;
	 ty /= norm;
	 dx = +tx*SPEED*scale;
	 dy = +ty*SPEED*scale;
	 flag = 1;
	 break;

       case CMD_FL:
/*	 printf("%d FL\n",i);/**/
	 turn = 1;
	 tx = state[i].dX*cos(-ToRad*TURN_RADIUS*scale) -
	    state[i].dY*sin(-ToRad*TURN_RADIUS*scale);
	 ty = state[i].dX*sin(-ToRad*TURN_RADIUS*scale) +
	    state[i].dY*cos(-ToRad*TURN_RADIUS*scale);
	 norm = sqrt(tx*tx+ty*ty);
	 tx /= norm;
	 ty /= norm;
	 dx = +tx*SPEED*scale;
	 dy = +ty*SPEED*scale;
	 flag = 1;
	 break;

       case CMD_RR:
/*	 printf("%d RR\n",i);/**/
	 turn = 1;
	 tx = state[i].dX*cos(-ToRad*TURN_RADIUS*scale) -
	    state[i].dY*sin(-ToRad*TURN_RADIUS*scale);
	 ty = state[i].dX*sin(-ToRad*TURN_RADIUS*scale) +
	    state[i].dY*cos(-ToRad*TURN_RADIUS*scale);
	 norm = sqrt(tx*tx+ty*ty);
	 tx /= norm;
	 ty /= norm;
	 dx = -tx*SPEED*scale;
	 dy = -ty*SPEED*scale;
	 flag = 1;
	 break;

       case CMD_RL:
/*	 printf("%d RL\n",i);/**/
	 turn = 1;
	 tx = state[i].dX*cos(ToRad*TURN_RADIUS*scale) -
	    state[i].dY*sin(ToRad*TURN_RADIUS*scale);
	 ty = state[i].dX*sin(ToRad*TURN_RADIUS*scale) +
	    state[i].dY*cos(ToRad*TURN_RADIUS*scale);
	 norm = sqrt(tx*tx+ty*ty);
	 tx /= norm;
	 ty /= norm;
	 dx = -tx*SPEED*scale;
	 dy = -ty*SPEED*scale;
	 flag = 1;
	 break;

       case CMD_R:
/*	 printf("%d R\n",i);/**/
	 dx = -state[i].dX*SPEED*scale;
	 dy = -state[i].dY*SPEED*scale;
	 flag = 1;
	 break;

       case CMD_STOP:
       default:
	 /* nop */
      }
      /*
       * if the robot did not move, leave it alone
       * if it would hit a wall, stop it
       */
      if(flag) {
	 flag = 1;
	 if(state[i].x + dx - ROBOT_RADIUS < WEST_WALL) {
	    flag = 0;
	 }
	 if(state[i].x + dx + ROBOT_RADIUS > EAST_WALL) {
	    flag = 0;
	 }
	 if(state[i].y + dy - ROBOT_RADIUS < SOUTH_WALL) {
	    flag = 0;
	 }
	 if(state[i].y + dy+ ROBOT_RADIUS > NORTH_WALL) {
	    flag = 0;
	 }
	 /*
	  * if it is outside a wall, move it inside
	  *	(this deploys a parked robot)
	  */
	 if(state[i].x - ROBOT_RADIUS < WEST_WALL) {
	    state[i].x = WEST_WALL + 1.4*ROBOT_RADIUS;/**/
	    state[i].dX = 0.0;
	    state[i].dY = 1.0;
	    flag = 0;
	 }
	 if(state[i].x + ROBOT_RADIUS > EAST_WALL) {
	    state[i].x = EAST_WALL - 1.4*ROBOT_RADIUS;/**/
	    state[i].dX = 0.0;
	    state[i].dY = 1.0;
	    flag = 0;
	 }
	 if(state[i].y  - ROBOT_RADIUS < SOUTH_WALL) {
	    state[i].y = SOUTH_WALL + 1.4*ROBOT_RADIUS;/**/
	    state[i].dX = 0.0;
	    state[i].dY = 1.0;
	    flag = 0;
	 }
	 if(state[i].y + ROBOT_RADIUS > NORTH_WALL) {
	    state[i].y = NORTH_WALL - 1.4*ROBOT_RADIUS;/**/
	    state[i].dX = 0.0;
	    state[i].dY = 1.0;
	    flag = 0;
	 }
	 if(flag) {	/* don't bother to do colision check if wall us */
	    for(j=0; j<MAX_ROBOTS; j++) {
	       if(i != j) {
		  double cx,cy;
		  cx = state[j].x - (state[i].x + dx);
		  cy = state[j].y - (state[i].y + dy);
		  if(cx*cx+cy*cy <= 4.0*ROBOT_RADIUS*ROBOT_RADIUS) {
		     printf("#%d ouch\n",i);
		     flag = 0;
		  }
	       }
	    }
	 }
	 /*
	  * if we are ordered to move and not colliding, then move!
	  */
	 if(flag) {
	    double rx,ry;
	    RandVector(SPEED*scale*ERR_SIZE, &rx, &ry);
	    state[i].x += (dx+rx);
	    state[i].y += (dy+ry);
	    if(turn) {
	       state[i].dX = tx;
	       state[i].dY = ty;
	    }
	 }
	 msg.robot = i;
	 msg.x = state[i].x;
	 msg.y = state[i].y;

	 coord.list[i].x = state[i].x;
	 coord.list[i].y = state[i].y;
      }
   }
   tcaInform(VisionInform, &coord);
}


/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void clockHnd(TCA_REF_PTR ref, ClockPtr time)
 *
 * DESCRIPTION:
 *	read the virtual clock broadcast and update the simulation
 *	appropriately
 *
\*--------------------------------------------------------------------------*/

int nextStep;	/* records the last known time	*/

void clockHnd(TCA_REF_PTR ref, ClockPtr time)
{
   while(nextStep < *time) {
      simulate(1000);
      nextStep += 1;
   }
   tcaFreeData(tcaReferenceName(ref), time);
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: long GetMilliSecs()
 *
 * DESCRIPTION:
 *	returns milliseconds since 1970
 *
\*--------------------------------------------------------------------------*/


long GetMilliSecs()
{
   struct timeval tval;

   if(gettimeofday(&tval, 0)) {
      fprintf(stderr, "gettimeofday() failed!!\n");
      exit(-1);
   }

   return tval.tv_sec*1000 + tval.tv_usec/1000;
}


/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: main()
 *
 * DESCRIPTION: 
 *
\*--------------------------------------------------------------------------*/

main()
{
   int i,j;
   double p,step;
   XYType msg;
   int width;
   long nextEvent;
   fd_set *listenPtr,listenList;


   nextStep = 0;	/* init time	*/

   /*
    * configure robots in parked position
    */
   step = 1.4*ROBOT_RADIUS;
   p = WEST_WALL + step;
   for(i=0; i<MAX_ROBOTS; i++,p+=step) {
      state[i].cmd = CMD_STOP;
      state[i].x = p;
      if(i%2) {
	 state[i].y = SOUTH_WALL - ROBOT_RADIUS;
      } else {
	 state[i].y = NORTH_WALL + ROBOT_RADIUS;
      }
      state[i].dX = 0.0;
      state[i].dY = 1.0;
   }

   /*
    * set up structure to send coords to vision sim
    */
   coordInitArray(&coord);
   coordGrowArray(MAX_ROBOTS, &coord);

   /*
    * clear robot cookies
    */
   for(i=0; i<MAX_ROBOTS; i++)
      cookie[i] = 0;
   newCookie = 476;

   printf("Connect...\n");

   tcaConnectModule(MODULE, tcaServerMachine());

   tcaWaitUntilReady();

   listenPtr = tcaGetConnections();
   width = ulimit(4, 256);

   tcaRegisterInformMessage(RobotCmd, RobotCmdFormat);/**/
   tcaRegisterHandler(RobotCmd, "robotCmdHnd", robotCmdHnd);

   tcaRegisterHandler(ClockBroadcast, "clockHnd", clockHnd);
   
   tcaRegisterQueryMessage(IncarnateQuery,
			   IncarnateQueryFormat,
			   IncarnateQueryReplyFmt);
   tcaRegisterHandler(IncarnateQuery, "incarnateQueryHnd", incarnateQueryHnd);

   /* send positions to vision simulator */
   for(i=0; i<MAX_ROBOTS; i++) {
      msg.robot = i;
      msg.x = state[i].x;
      msg.y = state[i].y;
      coord.list[i].x = state[i].x;
      coord.list[i].y = state[i].y;
   }
   tcaInform(VisionInform, &coord);

   tcaModuleListen();/**/

}
